from django.apps import AppConfig
from django.core.checks import Tags, register
from django.utils.translation import gettext_lazy as _

from ambient_toolbox.autodiscover.settings import get_autodiscover_enabled
from ambient_toolbox.static_role_permissions.settings import (
    get_static_role_permissions_enable_system_check,
    get_static_role_permissions_path,
)


class AmbientToolboxConfig(AppConfig):
    name = "ambient_toolbox"
    verbose_name = _("Ambient Toolbox")

    def ready(self):
        # System checks for static role permission feature
        if get_static_role_permissions_path() and get_static_role_permissions_enable_system_check():
            from ambient_toolbox.static_role_permissions.system_check import check_permissions_against_models

            register(check_permissions_against_models, Tags.models)

        # Function autodiscovery feature
        if get_autodiscover_enabled():
            # Register all decorated functions before they get imported by something else which will break the
            # registration process since decorators are only executed the first time.
            from ambient_toolbox.autodiscover import decorator_based_registry

            decorator_based_registry.autodiscover()



import json
from pathlib import Path
from unittest import mock

from django.core.cache import cache
from django.test import override_settings

from ambient_toolbox.autodiscover.registry import DecoratorBasedRegistry
from ambient_toolbox.autodiscover.settings import get_autodiscover_cache_key


def dummy_function(*args):
    return None


def dummy_function_2(*args):
    return None


def test_decorator_based_registry_init_regular():
    decorator_based_registry = DecoratorBasedRegistry()

    assert decorator_based_registry.registry == {}


def test_decorator_based_registry_singleton_works():
    decorator_based_registry_1 = DecoratorBasedRegistry()
    decorator_based_registry_1.registry[0] = "my_module"
    decorator_based_registry_2 = DecoratorBasedRegistry()

    assert decorator_based_registry_1 is decorator_based_registry_2
    assert decorator_based_registry_1.registry == decorator_based_registry_2.registry


def test_decorator_based_registry_register_regular():
    decorator_based_registry = DecoratorBasedRegistry()
    decorator = decorator_based_registry.register(registry_group="test")
    decorator(dummy_function)

    assert len(decorator_based_registry.registry) == 1
    assert "dummy_function" in str(decorator_based_registry.registry["test"][0])


def test_decorator_based_registry_register_second_function():
    decorator_based_registry = DecoratorBasedRegistry()
    decorator = decorator_based_registry.register(registry_group="test")
    decorator(dummy_function)
    decorator(dummy_function_2)

    assert len(decorator_based_registry.registry) == 1
    assert "dummy_function" in str(decorator_based_registry.registry["test"][0])
    assert "dummy_function_2" in str(decorator_based_registry.registry["test"][1])


def test_decorator_based_registry_register_two_groups():
    decorator_based_registry = DecoratorBasedRegistry()
    decorator = decorator_based_registry.register(registry_group="one")
    decorator(dummy_function)
    decorator = decorator_based_registry.register(registry_group="two")
    decorator(dummy_function_2)

    assert len(decorator_based_registry.registry) == 2  # noqa: PLR2004
    assert "dummy_function" in str(decorator_based_registry.registry["one"][0])
    assert "dummy_function_2" in str(decorator_based_registry.registry["two"][0])


def test_decorator_based_registry_autodiscover_target_is_python_module():
    cache.clear()

    decorator_based_registry = DecoratorBasedRegistry()
    decorator_based_registry.autodiscover(registry_group="autodiscover")

    # Assert two functions registered
    assert len(decorator_based_registry.registry) == 2  # noqa: PLR2004
    assert "testapp" in decorator_based_registry.registry.keys()
    assert "other" in decorator_based_registry.registry.keys()

    # Assert one function registered for "testapp"
    assert len(decorator_based_registry.registry["testapp"]) == 1
    assert {
        "module": "testapp.autodiscover.registered_functions",
        "name": "registered_dummy_function_testapp",
    } == decorator_based_registry.registry["testapp"][0]

    # Assert one function registered for "other"
    assert len(decorator_based_registry.registry["other"]) == 2  # noqa: PLR2004
    assert {
        "module": "testapp.autodiscover.registered_functions",
        "name": "registered_dummy_function_other",
    } == decorator_based_registry.registry["other"][0]
    assert {
        "module": "testapp.autodiscover.registered_functions",
        "name": "DummyClass",
    } == decorator_based_registry.registry["other"][1]


def test_decorator_based_registry_autodiscover_target_is_python_file():
    cache.clear()

    decorator_based_registry = DecoratorBasedRegistry()
    decorator_based_registry.autodiscover(registry_group="more_registered_functions")

    # Assert two functions registered
    assert len(decorator_based_registry.registry) == 1
    assert "no_module" in decorator_based_registry.registry.keys()

    # Assert one function registered for "testapp"
    assert len(decorator_based_registry.registry["no_module"]) == 1
    assert {
        "module": "testapp.more_registered_functions",
        "name": "even_more_registered_function",
    } == decorator_based_registry.registry["no_module"][0]


def test_decorator_based_registry_autodiscover_registry_group_contains_subpackages():
    cache.clear()

    decorator_based_registry = DecoratorBasedRegistry()
    decorator_based_registry.autodiscover(registry_group="handlers.commands")

    # Assert two functions registered
    assert len(decorator_based_registry.registry) == 1
    assert "commands" in decorator_based_registry.registry.keys()

    # Assert one function registered for "testapp"
    assert len(decorator_based_registry.registry["commands"]) == 1
    assert {
        "module": "testapp.handlers.commands.test_commands",
        "name": "my_command_handler",
    } == decorator_based_registry.registry["commands"][0]


@mock.patch("ambient_toolbox.autodiscover.registry.get_autodiscover_app_base_path", return_value=Path("/some/path"))
def test_decorator_based_registry_autodiscover_no_local_apps(*args):
    cache.clear()

    decorator_based_registry = DecoratorBasedRegistry()
    decorator_based_registry.autodiscover(registry_group="autodiscover")

    assert len(decorator_based_registry.registry) == 0


@mock.patch("importlib.import_module")
@mock.patch("importlib.reload")
def test_decorator_based_registry_autodiscover_caching_avoid_importing_again(
    mocked_reload_module, mocked_import_module
):
    cache.set(
        get_autodiscover_cache_key(),
        json.dumps({"testapp": ["dummy_function_testapp"], "other": ["dummy_function_other"]}),
    )

    decorator_based_registry = DecoratorBasedRegistry()
    decorator_based_registry.autodiscover(registry_group="autodiscover")

    assert mocked_reload_module.call_count == 0
    assert mocked_import_module.call_count == 0


def test_decorator_based_registry_autodiscover_load_handlers_from_cache_regular(*args):
    cache.set(
        get_autodiscover_cache_key(),
        json.dumps({"testapp": ["dummy_function_testapp"], "other": ["dummy_function_other"]}),
    )

    decorator_based_registry = DecoratorBasedRegistry()
    registered_callables = decorator_based_registry._load_handlers_from_cache()

    assert len(registered_callables) == 2  # noqa: PLR2004


@override_settings(CACHES={"default": {"BACKEND": "django.core.cache.backends.dummy.DummyCache"}})
def test_decorator_based_registry_autodiscover_load_handlers_from_cache_dummy_cache(*args):
    decorator_based_registry = DecoratorBasedRegistry()
    registered_callables = decorator_based_registry._load_handlers_from_cache()

    assert len(registered_callables) == 0


@override_settings(CACHES={"default": {"BACKEND": "django.core.cache.backends.dummy.DummyCache"}})
def test_get_registered_callables_found_and_executable():
    decorator_based_registry = DecoratorBasedRegistry()
    callables = decorator_based_registry.get_registered_callables(registry_group="autodiscover")

    assert len(callables) == 3  # noqa: PLR2004
    assert callables[0]() == "testapp"
    assert callables[1]() == "other"
    assert str(callables[2]()) == "DummyClass"



import logging

from ambient_toolbox.autodiscover.logger import get_logger
from ambient_toolbox.autodiscover.settings import get_autodiscover_logger_name


def test_get_logger():
    logger = get_logger()

    assert isinstance(logger, logging.Logger)
    assert logger.name == get_autodiscover_logger_name()



from ambient_toolbox.autodiscover.utils import unique_append_to_inner_list


def test_unique_append_to_inner_list_key_doesnt_exist():
    data = {}
    data = unique_append_to_inner_list(data=data, key="new_key", value=1)

    assert len(data) == 1
    assert "new_key" in data
    assert data["new_key"] == [1]


def test_unique_append_to_inner_list_key_exists():
    data = {"my_key": [1]}
    data = unique_append_to_inner_list(data=data, key="my_key", value=2)

    assert len(data) == 1
    assert "my_key" in data
    assert data["my_key"] == [1, 2]


def test_unique_append_to_inner_list_value_exists():
    data = {"my_key": [1]}
    data = unique_append_to_inner_list(data=data, key="my_key", value=1)

    assert len(data) == 1
    assert "my_key" in data
    assert data["my_key"] == [1]



from ambient_toolbox.autodiscover import decorator_based_registry


@decorator_based_registry.register(registry_group="testapp")
def registered_dummy_function_testapp():
    return "testapp"


@decorator_based_registry.register(registry_group="other")
def registered_dummy_function_other():
    return "other"


@decorator_based_registry.register(registry_group="other")
class DummyClass:
    def __str__(self) -> str:
        return "DummyClass"



import threading
from http import HTTPStatus
from unittest.mock import Mock

import pytest
from django.contrib.auth.models import AnonymousUser
from django.http import HttpResponse

from ambient_toolbox.middleware.current_user import CurrentUserMiddleware


@pytest.mark.filterwarnings("ignore:CurrentUserMiddleware is deprecated")
def test_current_user_is_none_if_no_user_given():
    assert CurrentUserMiddleware.get_current_user() is None


@pytest.mark.filterwarnings("ignore:CurrentUserMiddleware is deprecated")
def test_current_user_is_none_if_request_user_is_none():
    response = set_current_user(user=None)
    assert response.status_code == HTTPStatus.NO_CONTENT


@pytest.mark.filterwarnings("ignore:CurrentUserMiddleware is deprecated")
def test_current_user_is_same_as_request_user():
    new_user = Mock(user_name="test_user")
    response = set_current_user(user=new_user)
    assert response.status_code == HTTPStatus.OK


@pytest.mark.filterwarnings("ignore:CurrentUserMiddleware is deprecated")
def test_current_user_is_thread_safe():
    user1 = Mock(user_name="user1")
    user2 = Mock(user_name="user2")
    current_users = []
    ready_event = threading.Event()
    proceed_event = threading.Event()
    first_thread = threading.Thread(
        target=set_current_user,
        args=(user1, current_users, ready_event, proceed_event),
    )
    second_thread = threading.Thread(
        target=set_current_user,
        args=(user2, current_users),
    )
    first_thread.start()
    ready_event.wait()
    second_thread.start()
    second_thread.join()
    proceed_event.set()
    first_thread.join()

    assert current_users[0] == user2
    assert current_users[0].user_name == "user2"
    assert current_users[1] == user1
    assert current_users[1].user_name == "user1"


@pytest.mark.filterwarnings("ignore:CurrentUserMiddleware is deprecated")
def test_user_is_cleared_after_request():
    user = Mock(user_name="test_user")
    request = Mock(user=user)
    middleware = CurrentUserMiddleware(get_response=lambda request: HttpResponse(status=HTTPStatus.OK))
    response = middleware(request)

    assert response.status_code == HTTPStatus.OK
    assert CurrentUserMiddleware.get_current_user() is None


@pytest.mark.filterwarnings("ignore:CurrentUserMiddleware is deprecated")
def test_replaced_user_is_reflected_in_middleware():
    # This is roughly what happens when using eg. DRF token auth: during
    # middleware processing, request.user is an AnonymousUser, and at some
    # later point DRF replaces it with a user determined from given auth
    # info (eg. a token in the request header).
    # This should ideally be taken into account in our middleware since it
    # is also used to provide the user for `CommonInfo.lastmodified_by`.
    def get_response(request):
        replaced_user = Mock(user_name="replaced_user")
        request.user = replaced_user
        user_from_mw = CurrentUserMiddleware.get_current_user()
        if user_from_mw is not replaced_user:
            return HttpResponse(status=HTTPStatus.CONFLICT)  # pragma: no cover
        return HttpResponse(status=HTTPStatus.OK)

    middleware = CurrentUserMiddleware(get_response)
    anonymous_user = AnonymousUser()
    request = Mock(user=anonymous_user)
    response = middleware(request)

    assert response.status_code == HTTPStatus.OK


def set_current_user(user=None, current_users=None, ready_event=None, proceed_event=None):
    def get_response(request):
        if ready_event is not None:
            ready_event.set()
        if proceed_event is not None:
            proceed_event.wait()
        user_from_mw = CurrentUserMiddleware.get_current_user()
        if current_users is not None:
            current_users.append(user_from_mw)
        if user_from_mw is not request.user:
            return HttpResponse(status=HTTPStatus.CONFLICT)  # pragma: no cover
        if user_from_mw is None:
            return HttpResponse(status=HTTPStatus.NO_CONTENT)
        return HttpResponse(status=HTTPStatus.OK)

    middleware = CurrentUserMiddleware(get_response)
    request = Mock()
    request.user = user
    return middleware(request)



from pathlib import Path
from typing import Union

from django.conf import settings


def get_autodiscover_enabled() -> Union[Path, str]:
    """
    Switch to enable or disable function autodiscovery.
    Since the registry will be called on every request, we want to save overhead for projects that don't use it.
    """
    return getattr(settings, "AMBIENT_TOOLBOX_AUTODISCOVER_ENABLED", False)


def get_autodiscover_app_base_path() -> Union[Path, str]:
    """
    Base path of the application autodiscover should look for registered handlers.
    """
    return getattr(settings, "AMBIENT_TOOLBOX_AUTODISCOVER_APP_BASE_PATH", getattr(settings, "BASE_PATH", None))


def get_autodiscover_cache_key() -> str:
    """
    Cache key to store registered handlers in.
    """
    return getattr(settings, "AMBIENT_TOOLBOX_AUTODISCOVER_CACHE_KEY", "toolbox_autodiscover")


def get_autodiscover_logger_name() -> str:
    """
    Django logger name
    """
    return getattr(settings, "AMBIENT_TOOLBOX_AUTODISCOVER_LOGGER_NAME", "toolbox_autodiscover")



from typing import Union


def unique_append_to_inner_list(*, data: dict, key: Union[str, int], value) -> dict:
    """
    Inserts "value" in the dictionary "data" on "key".
    If "key" doesn't exist yet, it will create a new list containing "value".
    If "value" at "key" already exists, it won't be appended.
    """
    if key not in data:
        data[key] = [value]
    elif value not in data[key]:
        data[key].append(value)

    return data



import logging

from ambient_toolbox.autodiscover.settings import get_autodiscover_logger_name


def get_logger() -> logging.Logger:
    """
    Returns an instance of a toolbox Django logger
    """
    return logging.getLogger(get_autodiscover_logger_name())



from django.conf import settings
from django.test import override_settings

from ambient_toolbox.autodiscover.settings import (
    get_autodiscover_app_base_path,
    get_autodiscover_cache_key,
    get_autodiscover_enabled,
    get_autodiscover_logger_name,
)


@override_settings(AMBIENT_TOOLBOX_AUTODISCOVER_ENABLED=True)
def test_get_autodiscover_enabled_is_set():
    assert get_autodiscover_enabled() is True


def test_get_autodiscover_enabled_default_used():
    assert get_autodiscover_enabled() is False


@override_settings(AMBIENT_TOOLBOX_AUTODISCOVER_APP_BASE_PATH="/path/to/autodiscover")
def test_get_autodiscover_app_base_path_is_set():
    assert get_autodiscover_app_base_path() == "/path/to/autodiscover"


def test_get_autodiscover_app_base_path_default_used():
    assert get_autodiscover_app_base_path() == settings.BASE_PATH


@override_settings(AMBIENT_TOOLBOX_AUTODISCOVER_CACHE_KEY="new_cache_key")
def test_get_autodiscover_cache_key_is_set():
    assert get_autodiscover_cache_key() == "new_cache_key"


def test_get_autodiscover_cache_key_default_used():
    assert get_autodiscover_cache_key() == "toolbox_autodiscover"


@override_settings(AMBIENT_TOOLBOX_AUTODISCOVER_LOGGER_NAME="my_logger")
def test_get_autodiscover_logger_name_is_set():
    assert get_autodiscover_logger_name() == "my_logger"


def test_get_autodiscover_logger_name_default_used():
    assert get_autodiscover_logger_name() == "toolbox_autodiscover"
